'
' -----[ Title ]------------------------------------------------------
'
' File......: SNAP-006.BAS
' Purpose...: Air quality node for PLM-24
' Author....: Christer Johansson
' Version...: 1.02
' Stamp.....: BS1-IC
' Started...: 980503
' Updated...: 990125
'
' -----[ Program Description ]----------------------------------------
'
' This program shows how to implement the S.N.A.P protocol in a BS1-IC
' and use an gas sensor (AF30) from Scimarec to measure the air quialty
' in a room or building. The AF30 sensor is sensible for cigarette smoke
' but other variants are available for sensing Hydrogen, Carbon-monoxide,
' Methane, LP-Gas (propane, butane), Alcohol/Toluene/Gasoline vapour 
' and Ethanol/water. It uses PLM-24 Power Line Modems to communicate 
' over the power lines (mains). This example uses 16-bit CRC-CCITT as
' error detection method which gives secure data transfer.
'
' If the node is addressed by another node (PC or another MCU) it will
' return a value between $00-$FF in variable DB1 and a LED is toggled ON
' and OFF to indicate activity. The higher value the better air quality
' (i.e $E3=good air quality, $2A=you need to ventilate).
'
' The packet structure is defined in the received packets first two
' bytes (HDB2 and HDB1). The following packet structure is used.
'
' DD=01     - 1 Byte destination address
' SS=01     - 1 Byte source address
' PP=00     - No protocol specific flags
' AA=01     - Acknowledge is required
' C=0       - Command mode not supported
' EEE=100   - 16-bit CRC-CCITT
' NNNN=0001 - 1 Byte data
'
' Overview of header definition bytes (HDB2 and HDB1)
'
'        HDB2              HDB1
' +-----------------+-----------------+
' | D D S S P P A A | C E E E N N N N |
' +-----------------+-----------------+
'
' For more information about PLM-24 Power Line Modem see URL below.
'
' http://www.hth.com/plm-24/
'
' For more information about S.N.A.P see URL below.
'
' http://www.hth.com/snap/
'
' Below is an short overview of the S.N.A.P network protocol.
'
' Features
' --------
'
' - Free and open network protocol.
' - Easy to learn, use and implement.
' - Up to 16.7 million destination node addresses.
' - Up to 16.7 million source node addresses.
' - 0-24 protocol specific flags.
' - Optional ACK/NAK request.
' - Optional command mode.
' - 8 different error detecting methods.
' - Header is scaleable from 3-14 bytes.
' - 0-512 Bytes data in each packet.
' - Can be used as master/slave or peer-to-peer.
' - Supports broadcast messages.
'
' -----[ Disclaimer ]-------------------------------------------------
' 
' This example is offered on an "AS IS" basis, no warranty expressed
' or implied. HTH disclaim liability of any damages associated with
' the use of the hardware or software described herein. You use it on
' your own risk. We are not able to provide any free support.
'
' Copyright (c) 1998-1999 HTH all rights reserved
'
' -----[ Revision History ]-------------------------------------------
'
' 980503 - Version 1.00    Initial release
' 980911 - Verison 1.01    Revised for S.N.A.P Version 0.91
' 990125 - Version 1.02    Revised for S.N.A.P Version 1.00
'
' -----[ I/O Pin usage ]----------------------------------------------
' 
' Stamp pin	In/Out		External hardware	 
'
' Pin 7		  Out		LED
' Pin 6	  	  Out		AF30 gas sensor
' Pin 5		In              Not connected
' Pin 4		In              Not connected
' Pin 3	  	   Out 		PLM-24 Rx/Tx pin
' Pin 2	  	In		PLM-24 CD pin
' Pin 1	  	   Out		PLM-24 TxD pin 
' Pin 0	 	In		PLM-24 RxD pin
'
' -----[ Constants ]--------------------------------------------------
'
symbol Preamble		= %01010101	' Preamble byte	
symbol SYNC		= %01010100	' Synchronisation byte	
symbol RxD		= 0             ' PLM-24 Serial output pin
symbol TxD		= 1             ' PLM-24 Serial input pin
symbol CarrierDetect	= pin2          ' PLM-24 Carrier detect pin
symbol RxTx		= 3             ' PLM-24 Rx/Tx control pin
symbol LED		= 7		' LED pin
symbol AF30		= 6		' AF30 Sensor pin
symbol Baud		= T2400         ' Baudrate
symbol CRCPOLY		= $1021         ' CRC-CCITT
symbol MyAddress	= $03           ' Address for this node (1-255)
 		    
' -----[ Variables ]--------------------------------------------------
'
symbol CRC		= w0	' CRC Word
symbol Lo_CRC		= b0	' CRC Lo_Byte
symbol Hi_CRC		= b1	' CRC Hi_Byte
symbol DB1		= b2	' Packet data
symbol SAB1		= b3	' What node sent this packet
symbol DAB1		= b4	' What node should have this paket
symbol HDB1		= b5	' Header Definition Byte 1
symbol HDB2		= b6	' Header Definition Byte 2
symbol Byte		= b7	' Temporary Variable
symbol Temp		= b8	' Temporary Variable
symbol CRC2		= b9	' Packet CRC Hi_Byte
symbol CRC1		= b10   ' Packet CRC Lo_Byte

' -----[ Initialization ]---------------------------------------------
'
Pins = %10000010		' Set pin 7 and 1 high on startup
Dirs = %11001010                ' Set 7, 6, 3 and 1 as outputs

DB1 = 0				' Clear Data variable
High RxTx			' Set PLM-24 to receive mode

'------[ Program ]----------------------------------------------------
'

Start:
        ' Wait for preamble sync byte, if received read next seven
        ' bytes from master
        serin RxD,baud,(SYNC),HDB2,HDB1,DAB1,SAB1,DB1,CRC2,CRC1

        '
        ' Packet header check routine
        '
        ' Check HDB2 to see if STAMP are capable to use the packet
        ' structure, if not goto Start
        if HDB2 <> %01010001 then Start
        ' Check HDB1 to see if STAMP are capable to use the packet
        ' structure, if not goto Start
        if HDB1 <> %01000001 then Start

        '
        ' Address check routine
        '
        ' Check if this is the node addressed, if not goto Start
        if DAB1 <> MyAddress then Start

        '
        ' CRC check routine
        '
        ' Calculate CRC for all received bytes
        CRC = 0
        Byte = HDB2
        gosub calc_CRC
        Byte = HDB1
        gosub calc_CRC
        Byte = DAB1
        gosub calc_CRC
        Byte = SAB1
        gosub calc_CRC
        Byte = DB1
        gosub calc_CRC
        Byte = CRC2 
        gosub calc_CRC
        Byte = CRC1
        gosub calc_CRC
        ' Check if there was any CRC errors, if so send NAK
        if CRC <> 0 then NAK

        ' No CRC errors in packet
	DB1=0				' Clear Data variable
	low LED				' Turn LED on
	pot AF30,100,DB1		' Measure air quality
  	pause 100
        high LED                        ' Turn LED off 
        goto ACK                        ' Send ACK and return data
        
ACK:
        ' Send ACK (i.e tell master that packet was OK)
        ' Set ACK bits to ACK in HDB2 (logic XOR bit0 and bit1)
        HDB2 = HDB2 ^ %00000011
        goto send

NAK:
        ' Send NAK (i.e tell master that packet was bad)
        ' Set ACK bits to NAK in HDB2 (logic OR bit0 and bit1)
        HDB2 = HDB2 | %00000011
        goto send

Send:
        ' Swap SAB1 <-> DAB1 address bytes
        temp = SAB1
        SAB1 = DAB1
        DAB1 = temp

        ' Clear CRC variable
        CRC = 0
        ' Put HDB2 in variable Byte
        Byte = HDB2
        ' Calculate CRC
        gosub calc_crc
        
        ' Put HDB1 in variable Byte
        Byte = HDB1
        ' Calculate CRC
        gosub calc_crc
        
        ' Put DAB1 in variable Byte
        Byte = DAB1
        ' Calculate CRC
        gosub calc_crc

        ' Put SAB1 in variable Byte
        Byte = SAB1
        ' Calculate CRC
        gosub calc_crc

        ' Put DB1 in variable Byte
        Byte = DB1
        ' Calculate CRC
        Gosub calc_crc
        
        ' Move calculated Hi_CRC value to outgoing packet
        CRC2 = Hi_CRC
        ' Move calculated Lo_CRC value to outgoing packet
        CRC1 = Lo_CRC
        
Check_CD:       
        ' Check carrier detect pin if some other PLM-24 unit is sending       
        if CarrierDetect = 1 then Tx
        ' Someone else was sendning. Use current CRC value to generate
        ' a random pause time between 0-255 ms before trying again
        random CRC
        pause Lo_CRC
        goto Check_CD
Tx:
        ' Set PLM-24 Power Line Modem in transmit mode
        low RxTx
        ' Give PLM-24 time to settle before transmitt any data
        pause 50

        ' Send packet to master, including the preamble and sync byte
        serout TxD,baud,(Preamble,SYNC,HDB2,HDB1,DAB1,SAB1,DB1,CRC2,CRC1)

        ' Give STAMP time to shift out all bits before setting to Rx
        Pause 50
        ' Set PLM-24 Power Line Modem in receive mode
        high RxTx

        ' Done, go back to Start and wait for a new packet
        goto start
 
' -----[ Subroutines ]------------------------------------------------
'
' Subroutine for calculating CRC value in variable byte 
calc_crc:
        CRC = Byte * 256 ^ CRC
        for temp = 0 to 7
            if Bit15 = 0 then shift_only
            CRC = CRC * 2 ^ CRCPOLY
            goto nxt
shift_only: CRC = CRC * 2
nxt:    next
        return

' -----[ End ]--------------------------------------------------------

